(* Delphi 7 - Unit
   Extended media functions

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Apr. 2002
   Vers. 2 - Dec. 2005 - with volume control
   *)

unit Sound;

interface

uses Windows, MPlayer, mmSystem;

type
  TSoundFlag = (sfAsync,sfSync,sfLoop,sfNoStop,sfNoDefault);
  TSoundFlags = set of TSoundFlag;

  TMidiChannel = 0..$F;
  TMidiData = 0..$7F;
  TMidiShortMsg = packed record
  case Integer of
    0: (DWord: DWORD);
    1: (StatusByte: Byte;
        DataByte1,
        DataByte2: Byte);
    2: { Running status }
       (RDataByte1,
        RDataByte2: Byte);
  end;


const
  SoundID = 'WAVE';

var
  SoundAvailable : boolean;

{ ------------------------------------------------------------------- }
// Player
function PlaySoundFromFile (Filename : string;
                            Flags    : TSoundFlags) : boolean;

function PlaySoundFromMemory (ResID : string;
                              Flags : TSoundFlags) : boolean;

procedure StopSound;

{ ------------------------------------------------------------------- }
{ Volume: 0 - 1000 }
function GetMPVolume(MP: TMediaPlayer): Integer;
procedure SetMPVolume(MP: TMediaPlayer; Volume: Integer);

{ ------------------------------------------------------------------- }
//Midi
function OpenFMSynth(var MidiOut: hMidiOut; var Caps : TMidiOutCaps): Boolean;

procedure CloseFMSynth(MidiOut: hMidiOut);

procedure MidiNoteOn(Device   : hMidiOut;
                     Channel  : TMidiChannel;
                     Key      : TMidiData;
                     Velocity : TMidiData);

procedure MidiNoteOff(Device   : hMidiOut;
                      Channel  : TMidiChannel;
                      Key      : TMidiData;
                      Velocity : TMidiData);

{ ------------------------------------------------------------------- }
// Play frequency
procedure PlayFrequency (AFreq,ADuration : cardinal);

type
  TVolumeLevel = 0..127;

procedure MakeSound(Frequency{Hz}, Duration{mSec}: Integer; Volume: TVolumeLevel);

implementation

uses Classes;

{ ------------------------------------------------------------------- }
// Player
function GetSoundFlags (Flags : TSoundFlags) : cardinal;
begin
  Result:=0;
  if sfAsync in Flags then Result:=Result or SND_ASYNC;
  if sfSync in Flags then Result:=Result or SND_SYNC;
  if sfLoop in Flags then Result:=Result or SND_LOOP;
  if sfNoStop in Flags then Result:=Result or SND_NOSTOP;
  if sfNoDefault in Flags then Result:=Result or SND_NODEFAULT;
  end;

function PlaySoundFromFile (Filename : string;
                            Flags    : TSoundFlags) : boolean;
begin
  if SoundAvailable then Result:=sndPlaySound(pchar(Filename),GetSoundFlags(Flags))
  else Result:=false;
  end;

function PlaySoundFromMemory (ResID : string;
                              Flags : TSoundFlags) : boolean;
var
  Res,rh : THandle;
begin
  Result:=false;
  if SoundAvailable then begin
    Res:=FindResource(HInstance,pchar(ResID),SoundID);
    if Res<>0 then begin
      rh:=LoadResource(HInstance,Res);
      if rh<>0 then begin
        Result:=sndPlaySound(LockResource(rh),GetSoundFlags(Flags) or SND_MEMORY);
        UnlockResource(rh);
        FreeResource(rh);
        end;
      end;
    end;
  end;

procedure StopSound;
begin
  if SoundAvailable then sndPlaySound(nil,0);
  end;

{ ------------------------------------------------------------------- }
// Volume settings
const
  MCI_SETAUDIO = $0873;
  MCI_DGV_SETAUDIO_VOLUME = $4002;
  MCI_DGV_SETAUDIO_ITEM = $00800000;
  MCI_DGV_SETAUDIO_VALUE = $01000000;
  MCI_DGV_STATUS_VOLUME = $4019;

type
  TMCI_DGV_SETAUDIO_PARMS = record
    dwCallback : DWORD;
    dwItem     : DWORD;
    dwValue    : DWORD;
    dwOver     : DWORD;
    lpstrAlgorithm : PChar;
    lpstrQuality   : PChar;
  end;

type
  TMCI_STATUS_PARMS = record
    dwCallback : DWORD;
    dwReturn   : DWORD;
    dwItem     : DWORD;
    dwTrack    : DWORD;
  end;

procedure SetMPVolume(MP: TMediaPlayer; Volume: Integer);
  { Volume: 0 - 1000 }
var
  AudioParms : TMCI_DGV_SETAUDIO_PARMS;
begin
  { Volume: 0 - 1000 }
  with AudioParms do begin
    dwCallback:=0;
    dwItem:=MCI_DGV_SETAUDIO_VOLUME;
    dwValue:=Volume;
    dwOver:=0;
    lpstrAlgorithm:=nil;
    lpstrQuality:=nil;
    end;
  mciSendCommand(MP.DeviceID, MCI_SETAUDIO,
    MCI_DGV_SETAUDIO_VALUE or MCI_DGV_SETAUDIO_ITEM, Cardinal(@AudioParms));
  end;

function GetMPVolume(MP: TMediaPlayer): Integer;
var
  StatusParms : MCI_STATUS_PARMS;
begin
  with StatusParms do begin
    dwCallback:=0;
    dwItem:=MCI_DGV_STATUS_VOLUME;
    end;
  mciSendCommand(MP.DeviceID, MCI_STATUS, MCI_STATUS_ITEM, Cardinal(@StatusParms));
  Result:=StatusParms.dwReturn;
  { Volume: 0 - 1000 }
  end;


{ ------------------------------------------------------------------- }
// Midi synthesizer
const
  MidiDataMask = $7F;
  MidiChannelMask = $F;
  RunningStatus: Byte = 0;

  // the following constants are from mmsystem.h
  // they are missing in MMSystem
  MOD_MAPPER    = 5;   // MIDI mapper
  MOD_WAVETABLE = 6;   // hardware wavetable synth
  MOD_SWSYNTH   = 7;   // software synth

procedure MidiNoteOn(Device   : hMidiOut;
                     Channel  : TMidiChannel;
                     Key      : TMidiData;
                     Velocity : TMidiData);
var
  Msg: TMidiShortMsg;
begin
  with Msg do begin
    StatusByte:=$90 or Byte(Channel and MidiChannelMask);
    if StatusByte = RunningStatus then begin
      RDataByte1:=Key and MidiDataMask;
      RDataByte2:=Velocity and MidiDataMask;
      end
    else begin
      RunningStatus:=StatusByte;
      DataByte1:=Key and MidiDataMask;
      DataByte2:=Velocity and MidiDataMask;
      end;
    midiOutShortMsg(Device, DWord);
    end;
  end;

procedure MidiNoteOff(Device   : hMidiOut;
                      Channel  : TMidiChannel;
                      Key      : TMidiData;
                      Velocity : TMidiData);
var
  Msg: TMidiShortMsg;
begin
  with Msg do begin
    StatusByte:=$80 or Byte(Channel and MidiChannelMask);
    if StatusByte = RunningStatus then begin
      RDataByte1:=Key and MidiDataMask;
      RDataByte2:=Velocity and MidiDataMask;
      end
    else begin
      RunningStatus:=StatusByte;
      DataByte1:=Key and MidiDataMask;
      DataByte2:=Velocity and MidiDataMask;
      end;
    midiOutShortMsg(Device, DWord);
    end;
  end;

function OpenFMSynth(var MidiOut: hMidiOut; var Caps : TMidiOutCaps): Boolean;
var
  I    : Integer;
begin
  for I:=0 to midiOutGetNumDevs-1 do begin
    if (midiOutGetDevCaps(I,@Caps,SizeOf(Caps))=MMSYSERR_NOERROR)
        and ((Caps.wTechnology=MOD_FMSYNTH)
          or (Caps.wTechnology=MOD_SWSYNTH)) then begin
      Result:=midiOutOpen(@MidiOut, I, 0, 0, 0) = MMSYSERR_NOERROR;
      if Result then Exit;
      end;
    end;
  Result:=False;
  end;

procedure CloseFMSynth(MidiOut: hMidiOut);
begin
  midiOutClose(MidiOut);
  end;

{ ------------------------------------------------------------------- }
// Play frequency
procedure PlayFrequency (AFreq,ADuration : cardinal);
begin
  Beep(AFreq,ADuration);
  end;

procedure MakeSound(Frequency{Hz}, Duration{mSec}: Integer; Volume: TVolumeLevel);
  {writes tone to memory and plays it} 
var 
  WaveFormatEx: TWaveFormatEx; 
  i, TempInt, DataCount, RiffCount: integer;
  SoundValue: byte; 
  w: double; // omega ( 2 * pi * frequency) 
const 
  Mono: Word = $0001; 
  SampleRate: Integer = 11025; // 8000, 11025, 22050, or 44100 
  RiffId: string = 'RIFF'; 
  WaveId: string = 'WAVE'; 
  FmtId: string = 'fmt '; 
  DataId: string = 'data'; 
begin 
{  if Frequency > (0.6 * SampleRate) then
  begin 
    ShowMessage(Format('Sample rate of %d is too Low to play a tone of %dHz',
      [SampleRate, Frequency])); 
    Exit; 
  end;    }
  with WaveFormatEx do begin
    wFormatTag:=WAVE_FORMAT_PCM;
    nChannels:=Mono;
    nSamplesPerSec:=SampleRate;
    wBitsPerSample:=$0008;
    nBlockAlign:=(nChannels * wBitsPerSample) div 8;
    nAvgBytesPerSec:=nSamplesPerSec * nBlockAlign;
    cbSize:=0;
    end;
  with TMemoryStream.Create do begin
    {Calculate length of sound data and of file data}
    DataCount:=(Duration*SampleRate) div 1000; // sound data
    RiffCount:=Length(WaveId)+Length(FmtId)+SizeOf(DWORD)+
      SizeOf(TWaveFormatEx)+Length(DataId)+SizeOf(DWORD)+DataCount; // file data
    {write out the wave header}
    Write(RiffId[1],4); // 'RIFF'
    Write(RiffCount,SizeOf(DWORD)); // file data size
    Write(WaveId[1],Length(WaveId)); // 'WAVE'
    Write(FmtId[1],Length(FmtId)); // 'fmt '
    TempInt:=SizeOf(TWaveFormatEx);
    Write(TempInt,SizeOf(DWORD)); // TWaveFormat data size
    Write(WaveFormatEx,SizeOf(TWaveFormatEx)); // WaveFormatEx record
    Write(DataId[1],Length(DataId)); // 'data'
    Write(DataCount,SizeOf(DWORD)); // sound data size
    {calculate and write out the tone signal} // now the data values
    w:=2*Pi*Frequency; // omega
    for i:=0 to DataCount-1 do begin
      SoundValue:=127+trunc(Volume*sin(i*w/SampleRate)); // wt = w * i / SampleRate
      Write(SoundValue,SizeOf(Byte));
      end;
    {now play the sound}
    sndPlaySound(Memory, SND_MEMORY or SND_ASYNC);
    Free;
  end;
end;

{ ------------------------------------------------------------------- }
begin
  SoundAvailable:=WaveOutGetNumDevs>0;
  end.
