(* Mathematische Routinen zur Approximation von Messwerten
   =======================================================
   
    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jun. 2001 *)

unit WinFit;

interface

const
  GradMax = 15;
  MatMax = 2*GradMax;
  IterMax   : integer = 50;
  Precision : double = 1E-15;    (* Determinantengrenze fr die Iterationen *)

type
  TMatrix = array [1..MatMax,1..GradMax] of double;
  TFloatVektor = array [1..MatMax] of double;
  TIntVektor = array [1..MatMax] of integer;
  TPolKoeff  = array [0..GradMax] of double;
  SpKoeff = array [1..10,1..3] of double;
  PFloatVektor  = ^TFloatVektor;
  PIntVektor  = ^TIntVektor;

function Horner (const A;
                 N  : integer;
                 X  : double) : double;

procedure LoesGl (const SMAT;
                  const B;
                  Dim,N    : integer;
                  var A;
                  var DET  : double;
                  var ITER : integer);

function PolFit (Grad,n : integer;
                 const x,y;
                 var PolKoeff : TPolKoeff;
                 var Sigma    : double) : boolean;

function CubSpline (NP,Smooth : integer;
                    const XA,YA;
                    var YS,CK) : integer;
                    
function CubSplineVal (NP     : integer;
                       X      : double;
                       const XA,YS,CK;
                       var Y     : double) : integer;

function CubSplineFit (NP,Smooth : integer;
                       const XA,YA) : integer;

function CubSplineFitValue (x : double) : double;


implementation
{------------------------------------------------------------------}
(* Berechnung eines Polynoms nach dem Hornerschema
   A = Polynomkoeffizienten
   N = Grad des Polynoms
   X = Wert
   HORNER = Pol(X)
   J. Rathlev, Mrz 1988 *)
function Horner (const A;
                 N  : integer;
                 X  : double) : double;
var
  I  : integer;
  Y  : double;
begin
  if (X <> 0) then begin
    Y:=TPolKoeff(A)[N];
    for I:=pred(N) downto 0 do Y:=Y*X+TPolKoeff(A)[I];
    HORNER := Y;
    end
  else HORNER := TPolKoeff(A)[0];
  end;

{------------------------------------------------------------------}
(* Unterroutine zum Loesen von linearen Gleichungssystemen
   nach dem Couffignal Verfahren
       Uebergabeparameter:
       SMAT = Systemmatrix
              (Zeilenindex,Spaltenindex)
       B = rechte Seite        (SMAT und B werden nicht veraendert)
       Dim = Dimension des Arrays SMAT
       N = Ordnung des Systems
       A = Loesungsvektor
       DET = Determinante
       ITER = Anzahl der Iterationen zur Reduzierung der Rundungsfehler
            =  0 ==> Die Determinante ist zu klein (d.h. die Loesung
                     ist unkorrigiert)
            = -1 ==> Die Determinante verschwindet, keine Lsung
   Die Ordnung des Gleichungssystem ist beliebig (<=25)*)
procedure LoesGl (const SMAT;
                  const B;
                  Dim,N    : integer;
                  var A;
                  var DET  : double;
                  var ITER : integer);

  function AMAX (x,y : double) : double;
  begin
    if x<y then amax:=y else amax:=x;
    end;

var
  M            : PFloatVektor;
  KS,KZ        : PIntVektor;
  Q0,Q1,Q2     : double;
  I0,I1,I2,N1,
  J0,J1,J2,N3  : integer;
  ende         : boolean;

begin
  Iter:=-1;
  N3:=N+2; DET := 1.0;
  GetMem (M,N3*N*SizeOf(double));
  GetMem (KS,N*SizeOf(integer));
  GetMem (KZ,N*SizeOf(integer));
(* UMSPEICHERN IN DIE MATRIX M *)
  for I1:=1 to N do
    for I2:=1 to N do M^[I1+N*pred(I2)]:=TFloatVektor(SMAT)[Dim*pred(I1)+I2];
  N1:=N*N;
  for I1:=1 to N do M^[I1+N1]:=-TFloatVektor(B)[I1];
  for I1:=1 to N do begin
    KS^[I1]:=0; KZ^[I1]:=0;
    end;
  for I0:=1 to N do begin
 (* ERMITTLUNG DES PIVOT-ELEMENTS *)
    Q0:=2.0;
    for I1:=1 to N do
      IF (KZ^[I1] = 0) then begin
        Q2:=0.0;
        for I2:=1 to N do
          IF (KS^[I2] = 0) then begin
            Q1:=ABS(M^[I1+N*pred(I2)]);
            IF (Q1 > Q2) then begin
              Q2:=Q1; J0:=I2;
              end;
            end;
        Q1:=0.0;
        ende:=abs(Q2)<Precision;
        if ende then break; (* Pivot Elem. zu klein *)
        for I2:=1 to N do
          IF (KS^[I2] = 0) and (J0 <> I2) then
            Q1:=AMAX(Q1,ABS(M^[I1+N*pred(I2)]/Q2));
        IF (Q0 > Q1) then begin
          Q0:=Q1; J1:=I1; J2:=J0;
          end;
        end;
    if ende then break;
    KZ^[J1] := J2;            (* J1 ZEILENNUMMERN *)
    KS^[J2] := J1;            (* J2 SPALTENNUMMERN *)
    Q0:=M^[J1+N*pred(J2)];             (* Pivotelement *)
  (* LOESEN DES GLEICHUNGSSYSTEMS *)
    DET := DET*Q0;            (* DETERMINANTE *)
  (* SETZEN DER SPALTE N+2 FUER DIE INV. MATR. *)
    N1:=N*(N3-1);
    for I1:=1 to N do M^[I1+N1]:=0.0;
    M^[J1+N1]:=-1.0;
  (* Zeile J1 bearbeiten *)
    for I1:=1 to N3 do M^[J1+N*pred(I1)] := -M^[J1+N*pred(I1)]/Q0;
  (* Spalte J2 eliminieren *)
    for I1:=1 to N do
      IF (I1 <> J1) then begin
        Q0 := M^[I1+N*pred(J2)];
        for I2:=1 to N3 do
          M^[I1+N*pred(I2)] := M^[I1+N*pred(I2)]+Q0*M^[J1+N*pred(I2)];
        end;
  (* UMSPEICHERN DER SPALTE N+2 NACH J2 *)
    for I1:=1 to N do M^[I1+N*pred(J2)]:=M^[I1+N1];
    end;
  if not ende then begin
  (* ORDNEN DER LOESUNG *)
    N1:=N*N;
    for I1:=1 to N do TFloatVektor(A)[KZ^[I1]] := M^[I1+N1];
    IF (ABS(DET) > precision) then begin
    (* KORREKTUR DER RUNDUNGSFEHLER *)
      N1:=N*pred(N3); I0:=0;
      repeat
        for I1:=1 to N do begin
          Q0:=0.0;
          for I2:=1 to N do Q0:=Q0+TFloatVektor(SMAT)[Dim*pred(I1)+I2]*TFloatVektor(A)[I2];
          M^[I1+N1]:=Q0-TFloatVektor(B)[I1];
          end;
        Q1:=0.0;
        for I1:=1 to N do begin
          Q0:=0.0;
          for I2:=1 to N do Q0:=Q0+M^[I1+N*pred(I2)]*M^[KS^[I2]+N1];
          J0:=KZ^[I1];
          TFloatVektor(A)[J0]:=TFloatVektor(A)[J0]-Q0;          (* Q0 = KORREKTUR *)
          if Q0<>0 then Q1:=AMAX(Q1,ABS(Q0/TFloatVektor(A)[J0]));
          end;
        inc(I0); ITER := I0;
        ende:=abs(Q1)<=precision;
        until (I0>IterMax) or ende;
      end
    else ITER:=0;
    end
  else ITER:=-1;
  FreeMem (M,N3*N*SizeOf(double));
  FreeMem (KS,N*SizeOf(integer));
  FreeMem (KZ,N*SizeOf(integer));
  end;

{------------------------------------------------------------------}
(* Approximation mit Polynom nach Gauss *)
function PolFit (Grad,n : integer;
                 const x,y;
                 var PolKoeff : TPolKoeff;
                 var Sigma    : double) : boolean;
var
  tx,sx,sy,mx,my : double;
  NVek           : array [0..MatMax] of double;
  i,j            : integer;
  SMat           : TMatrix;
  BVek,LVek      : TFloatVektor;
begin
  if (n>Grad) and (Grad<GradMax) then begin
    (* Mittelwert berechnen *)
    sx:=0; sy:=0;
    for i:=1 to n do begin
      sx:=sx+TFloatVektor(x)[i];
      sy:=sy+TFloatVektor(y)[i];
      end;
    mx:=sx/n; my:=sy/n;
    if mx=0 then mx:=1; // keine Skalierung
    if my=0 then my:=1;
   (* Gau-Matrix berechnen *)
    for i:=1 to Grad+1 do BVek[i]:=0;
    for i:=1 to 2*Grad do NVek[i]:=0;
    for i:=1 to n do begin
      sx:=TFloatVektor(x)[i]/mx;
      sy:=TFloatVektor(y)[i]/my;
      tx:=sx;
      for j:=1 to 2*GradMax do begin
        NVek[j]:=NVek[j]+sx;
        if j<=Grad+1 then BVek[j]:=BVek[j]+sy;
        sx:=sx*tx; sy:=sy*tx;
        end;
      end;
    NVek[0]:=n;
    for i:=1 to Grad+1 do begin
      for j:=1 to i do SMat[j,i+1-j]:=NVek[i-1];
      If i<>1 then
        for j:=i to Grad+1 do SMat[j,Grad+1+i-j]:=NVek[Grad+i-1];
      end;
    (* Gleichungssystem lsen *)
    LoesGl (SMat,BVek,GradMax,Grad+1,LVek,sx,i);
    if i>=0 then begin
      (* Normierung zurckrechnen *)
      tx:=1;
      for i:=0 to Grad do begin
        PolKoeff[i]:=LVek[i+1]*my/tx; tx:=tx*mx;
        end;
      (* Berechnung der Fehlerquadrate *)
      sx:=0;
      for i:=1 to n do
        sx:=sx+sqr(TFloatVektor(y)[i]-Horner(PolKoeff,Grad,TFloatVektor(x)[i]));
      Sigma:=sqrt(sx/pred(n)/n);
      Result:=true;
      end
    else Result:=false;
    end
  else Result:=false;
  end;

{------------------------------------------------------------------}
(* Datenglttung mit kub. Spline Funktionen
   Basis:  FH KIEL MATHLIB  Rel. 1.1   01.03.91
           (C) Copyright 1991 by Prof. Dr. Guido Hartmann, Kiel
   Pascal-Umsetzung: J. Rathlev, Jun. 2001
   *)

type
  TSpWork = array [1..7,0..10] of double;
  PSpWork = ^TSpWork;

var
  NPoints     : integer;
  pck         : ^SpKoeff;
  pxa,pys     : ^TFloatVektor;

{------------------------------------------------------------------}
(* Purpose:   Data smoothing by a cubic spline.
   Parameters:
     NP:       (In)  The number of elements in XA and YA. NP must not
                     be less than 2.
     SMooth:   (In)  Smoothing parameter. SM is a measure for the sum
                     of squares of the differences between the input
                     values YA[i] and the values of the spline function
                     YS[i] at XA[i]. SM is normalized such that
                     SM = 1000  means maximum smoothing (linear fit) and
                     SM = 0    means interpolation.
                     Therefore SM must neither be less than zero nor
                     greater than hundred.
     XA:       (In)  Vector of length NP containing the abscissas of
                     the data points. XA must be ordered so that
                     XA[i] < XA[i+1].
     YA:       (In)  Vector of length NP containing the ordinates of
                     the data points.
     YS:       (Out) Vector of length NP containing the function values
                     of the spline corresponding to the abscissas XA.
     CK:       (Out) Coeff. matrix of length (NP-1) containing the spline
                     coefficients. The value of the spline approximation
                     at X is
                       S(X) = ((CK[I,3]*D+CK[I,2])*D+CK[I,1])*D+YS[i],
                     where XA[i]<=X<=XA[i+1], D=X-XA[i] and
  Result:      (Out) Error code
    Possible Errors
      0:  no error
      81: N is less than 2
      82: Input abscissas are not ordered
      83: SM is out of range
  Remarks:
    For evaluation of the spline function, program CubSplineVal may be used.
    Note that vector YS instead of YA has to be passed to CubSplineVal.
  Ref.:      C.H. Reinsch, Smoothing by Spline Functions
             Numerische Mathematik 10 (1967) 177-183
*)
function CubSpline (NP,Smooth : integer;
                    const XA,YA;
                    var YS,CK) : integer;
var
  Work         : array of array of double;
  e,f,f2,g,h,
  p,s,xsm,
  xi,yi,sx,sy,
  sxx,sxy,syy  : double;
  m2,ier,i     : integer;
  EndLoop      : boolean;
begin
(* TEST FOR PARAMETER ERRORS *)
  CubSpline:=81;
  IF NP<=1 then Exit;
  CubSpline:=82;
  for I:=2 to NP do if TFloatVektor(xa)[i]<=TFloatVektor(xa)[i-1] then Exit;
  CubSpline:=83;
  IF (Smooth<0) or (Smooth>100) then Exit;
  SetLength (Work,8,NP+2);
//  GetMem (Work,7*(NP+2)*SizeOf(double));
  xsm:=Smooth/1000;
  CubSpline:=0;
(* COMPUTE DEVIATION FROM STRAIGHT LINE FIT *)
  SX:=0; SY:=0;
  for I:=1 to NP do begin
    SX:=SX+TFloatVektor(xa)[i]; SY:=SY+TFloatVektor(ya)[i];
    end;
  SX:=SX/NP; SY:=SY/NP;
  SXX:=0; SXY:=0; SYY:=0;
  for I:=1 to NP do begin
    XI:=TFloatVektor(xa)[i]-SX; YI:=TFloatVektor(ya)[i]-SY;
    SXX:=SXX+XI*XI; SXY:=SXY+XI*YI; SYY:=SYY+YI*YI;
    end;
  S:=(SYY-SXY*SXY/SXX)*xsm;
(* INITIALIZATIONS *)
  M2:=NP+1;
  Work[1,0]:=0; Work[1,1]:=0;
  Work[2,NP]:=0; Work[3,NP]:=0;
  Work[3,M2]:=0; Work[6,0]:=0;
  Work[6,1]:=0; Work[6,NP]:=0; Work[6,M2]:=0;
  M2:=NP-1;
  H:=TFloatVektor(xa)[2]-TFloatVektor(xa)[1]; F:=(TFloatVektor(ya)[2]-TFloatVektor(ya)[1])/H;
(* COMPUTE MATRICES T AND Q *)
  for I:=2 to M2 do begin
    G:=H; H:=TFloatVektor(xa)[i+1]-TFloatVektor(xa)[i];
    E:=F; F:=(TFloatVektor(ya)[i+1]-TFloatVektor(ya)[i])/H;
    TFloatVektor(ys)[i]:=F-E;
    Work[4,I]:=(G+H)*2/3;
    Work[5,I]:=H/3;
    Work[1,I]:=1/H;
    Work[3,I]:=1/G;
    Work[2,I]:=-(1/G+1/H);
    end;
  for I:=2 to M2 do begin
    SpKoeff(CK)[I,1]:=Work[1,I]*Work[1,I]+Work[2,I]*Work[2,I]+Work[3,I]*Work[3,I];
    SpKoeff(CK)[I,2]:=Work[1,I]*Work[2,I+1]+Work[2,I]*Work[3,I+1];
    SpKoeff(CK)[I,3]:=Work[1,I]*Work[3,I+2];
    end;
  F2:=-S; P:=0;
(* LOOP TO DETERMINE THE LAGRANGIAN PARAMETER P *)
  repeat
(* CHOLESKY DECOMPOSITION *)
    for I:=2 to M2 do begin
      Work[2,I-1]:=Work[1,I-1]*F;
      Work[3,I-2]:=Work[1,I-2]*G;
      Work[1,I]:=1/(SpKoeff(CK)[I,1]*P+Work[4,I]-Work[2,I-1]*F-Work[3,I-2]*G);
      Work[6,I]:=TFloatVektor(ys)[i]-Work[2,I-1]*Work[6,I-1]-Work[3,I-2]*Work[6,I-2];
      F:=SpKoeff(CK)[I,2]*P+Work[5,I]-Work[2,I-1]*H;
      G:=H; H:=SpKoeff(CK)[I,3]*P;
      end;
(* COMPUTE VECTOR U *)
    for I:=M2 downto 2 do Work[6,I]:=Work[1,I]*Work[6,I]-
      Work[2,I]*Work[6,I+1]-Work[3,I]*Work[6,I+2];
    E:=0; H:=0;
(* COMPUTE VECTOR V *)
    for I:=1 to M2 do begin
      G:=H; H:=(Work[6,I+1]-Work[6,I])/(TFloatVektor(xa)[i+1]-TFloatVektor(xa)[i]);
      Work[7,I]:=H-G; E:=E+Work[7,I]*(H-G);
      end;
    G:=-H; Work[7,NP]:=G;
    E:=E-G*H; G:=F2; F2:=E*P*P;
    EndLoop:=true;
    IF (F2<S) and (F2>G) then begin
      F:=0;
      H:=(Work[7,2]-Work[7,1])/(TFloatVektor(xa)[2]-TFloatVektor(xa)[1]);
      for I:=2 to M2 do begin
        G:=H;
        H:=(Work[7,I+1]-Work[7,I])/(TFloatVektor(xa)[i+1]-TFloatVektor(xa)[i]);
        G:=H-G-Work[2,I-1]*Work[1,I-1]-Work[3,I-2]*Work[1,I-2];
        F:=F+Work[1,I]*G*G;
        Work[1,I]:=G;
        end;
      H:=E-P*F;
      IF (H>0) then begin
        (* ADJUST P AND ITERATE *)
        P:=P+(S-F2)/((SQRT(S/E)+P)*H);
        EndLoop:=false;
        end;
      end;
    until EndLoop;
(* DETERMINATION OF THE SPLINE *)
(* COEFFICIENTS *)
  for I:=1 to M2 do begin
    TFloatVektor(ys)[i]:=TFloatVektor(ya)[i]-Work[7,I]*P;
    SpKoeff(CK)[I,2]:=Work[6,I];
    end;
  TFloatVektor(ys)[NP]:=TFloatVektor(ya)[NP]-Work[7,NP]*P;
  for I:=1 to M2 do begin
    H:=TFloatVektor(xa)[i+1]-TFloatVektor(xa)[i];
    SpKoeff(CK)[I,3]:=(Work[6,I+1]-Work[6,I])/(H*3);
    SpKoeff(CK)[I,1]:=(TFloatVektor(ys)[i+1]-TFloatVektor(ys)[i])/H-(SpKoeff(CK)[I,3]*H+Work[6,I])*H;
    end;
//  FreeMem (Work,7*(NP+2)*SizeOf(double));
//  FreeMem (Work);
  end;

{------------------------------------------------------------------}
(* Compute y-value of spline interpolation
   Parameters:
     NP:       (In)  The number of elements in XA and YA. NP must not
                     be less than 2.
     X:        (In)  abscissa value of point to interpolate
     XA:       (In)  Vector of length NP containing the abscissas of
                     the data points. XA must be ordered so that
                     XA[i] < XA[i+1].
     YS:       (In)  Vector of length NP containing the function values
                     of the spline corresponding to the abscissas XA.
     CK:       (In)  Coeff. matrix of length (NP-1) containing the spline
                     coefficients. The value of the spline approximation
                     at X is
                       S(X) = ((CK[I,3]*DX+CK[I,2])*DX+CK[I,1])*DX+YS[i],
                     where XA[i]<=X<=XA[i+1], DX=X-XA[i] and
  Result:      (Out) Error code
    Possible Errors
      0:  no error
      84: X<XMin
      85: X>XMax
  *)
function CubSplineVal (NP     : integer;
                       X      : double;
                       const XA,YS,CK;
                       var Y     : double) : integer;
var
  i  : integer;
  dx : double;
begin
  (* Seek interval *)
  CubSplineVal:=84;
  if (x<TFloatVektor(XA)[1]) then Exit;
  CubSplineVal:=85;
  if (x>TFloatVektor(XA)[NP]) then Exit;
  CubSplineVal:=0;
  i:=2;
  while (x>=TFloatVektor(XA)[i]) and (i<NP) do inc(i);
  dec(i); dx:=x-TFloatVektor(XA)[i];
  Y:=((SpKoeff(CK)[I,3]*dx+SpKoeff(CK)[I,2])*dx+SpKoeff(CK)[I,1])*dx+TFloatVektor(YS)[i];
  end;

{------------------------------------------------------------------}
(* Data smoothing by a cubic spline.
   All results are stored to compute a value with GetCubSpline *)
function CubSplineFit (NP,Smooth : integer;
                       const XA,YA) : integer;
var
  i : integer;
begin
  if pxa<>nil then FreeMem (pxa,NPoints*SizeOf(double));
  GetMem (pxa,NP*SizeOf(double));
  if pys<>nil then FreeMem (pys,NPoints*SizeOf(double));
  GetMem (pys,NP*SizeOf(double));
  if pck<>nil then FreeMem (pck,3*(NPoints-1)*SizeOf(double));
  GetMem (pck,3*(NP-1)*SizeOf(double));
  NPoints:=NP;
  for i:=1 to NPoints do pxa^[i]:=TFloatVektor(xa)[i];
  Result:=CubSpline (NPoints,Smooth,pxa^,ya,pys^,pck^);
  end;

{------------------------------------------------------------------}
(* Compute interpolted value, no range cheking is done *)
function CubSplineFitValue (x : double) : double;
var
  i : integer;
  y : double;
begin
  if NPoints>1 then begin
    i:=CubSplineVal (NPoints,x,pxa^,pys^,pck^,y);
    if i=84 then Result:=pys^[1]
    else if i=85 then Result:=pys^[NPoints]
    else Result:=y;
    end
  else Result:=0;
  end;

begin
  NPoints:=0;
  end.
