(* Delphi Unit
   several mathematical functions
   ==============================

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jun. 1989
     last changed: Nov. 2001 
   *)

unit WinMath;

interface

uses SysUtils;

type
  EinhStr = string[5];

const
  Pi180 = Pi/180.0;   (* Grad -> Bogenma *)
  Pi200 = Pi/200.0;   (* Neugrad -> Bogenma *)
  MinDouble : double = -1E300;
  MaxDouble : double = 1E300;

var
  ln2,ln10,sqrt2 : extended;

function Pwr2 (exp : integer) : int64;

function AInt (x : double) : double;

function RMod (x,y : double) : double;

function Sign (x,y : double) : double;

function Lg (x : double) : double;

function Ld (x : double) : double;

function Pwr (x,y : double) : double;

function Kub (x : double) : double;

function Tan (x : double) : double;

function ArcTan1 (x,y : double) : double;

function ArcTan2 (x,y : double) : double;

function ArcTanB (x,y : double) : double;

function ISign (x,y : integer) : integer;

function RMin (x,y : double) : double;

function RMax (x,y : double) : double;

function NumberOfDigits (n : integer) : integer;

(* Typumwandlung prfen *)
function CheckValue (s     : string) : boolean;

(* an Dezimalstelle runden *)
function DecimalRound (Value : double; Decimal : integer) : double;

(* Normalisieren eines Wertes in den Bereich 0.1 .. 999.9 *)
procedure Normal (var Value,Factor : double;
                  var PrefixIndex  : integer);

(* Zahlenwert mit nachfolgendem Einheitenvorsatz erzeugen *)
function PrefixVal (Value : double;
                    n,nd  : integer;
                    Einh  : EinhStr) : string;

(* Zahlenwert aus String mit nachfolgendem Einheitenvorsatz erzeugen *)
function ValPrefix (ValStr    : string;
                    var Value : double) : boolean;

(* Zahlen auf "Eng-Format" umsetzen *)
function FloatToStrP (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;

(* Zahlen auf "Eng-Format" umsetzen *)
function FloatToPlotStr (Value   : double;
                         n       : integer) : string;

(* Zahlen auf "Eng-Format" mit Vorstzen (m,k,..)  umsetzen *)
function FloatToPrefixStr (Value     : double;
                           n         : integer;
                           Separator : string) : string;

(* Zahlen auf "ffXixed-Format" umsetzen *)
function FloatToFixStr (Value     : double;
                        n         : integer) : string;
                           
implementation
const
  PfAnz = 10;
  Pfx   : array [1..PfAnz] of char = ('f','p','n','','m',' ','k','M','G','T');
  PfFak : array [0..PfAnz] of double = (1.0,1E15,1E12,1E9,1E6,1E3,
                                        1.0,1E-3,1E-6,1E-9,1E-12);
  FFak : array [-2..8] of double = (0.01,0.1,1.0,10.0,100.0,1000.0,1E4,1E5,1E6,1E7,1E8);
  NFak : array [1..2] of integer = (10,100);

{ ------------------------------------------------------------------- }
(* Compute power of 2 as integer *)
function Pwr2 (exp : integer) : int64;
begin
  if exp>63 then exp:=63;
  if exp<0 then exp:=0;
  Result:=1;
  Result:=Result shl exp;
//  for i:=1 to exp do Result:=2*Result;
  end;


(* ----------FUNCTION--AINT--------------------------IEAP-KIEL----------
   Bestimmung des nchstkleineren ganzzahligen Anteils einer Zahl
*)
function AInt (x : double) : double;
var
  xx  : double;
begin
  xx:=int(x);
  if (x<0.0) and (x<>xx) then xx:=xx-1.0;
  aint:=xx;
  end;

(* ----------FUNCTION--RMOD--------------------------IEAP-KIEL----------
   Divisionsrest bei Fliekommazahlen
*)
function RMod (x,y : double) : double;
begin
  RMod:=x-int(1.000001*x/y)*y;  // Faktor erforderlich wegen Rundungsfehler
  end;

(* ----------FUNCTION--ISIGN-------------------------IEAP-KIEL----------
  (Vorzeichen von arg2) * arg1
*)
function ISign (x,y : integer) : integer;
begin
  if y<0 then isign:=-x else isign:=x; 
  end;

(* ----------FUNCTION--SIGN--------------------------IEAP-KIEL----------
  (Vorzeichen von arg2) * arg1
*)
function Sign (x,y : double) : double;
begin
  if y<0.0 then sign:=-x else if y>0.0 then sign:=x else sign:=0.0;
  end;

(* ----------FUNCTION--LG----------------------------IEAP-KIEL----------
  Dekadischer Logarithmus
*)
function Lg (x : double) : double;
begin
  lg:=ln(x)/ln10;
  end;

(* ----------FUNCTION--Ld----------------------------IEAP-KIEL----------
  Dualer Logarithmus
*)
function Ld (x : double) : double;
begin
  Ld:=ln(x)/ln2;
  end;

(* ----------FUNCTION--PWR---------------------------IEAP-KIEL----------
  allgemeine Exponentialfunktion x ^ y
*)
function Pwr (x,y : double) : double;
begin
  pwr:=exp (y*ln(x));
  end;

(* ----------FUNCTION--KUB---------------------------IEAP-KIEL----------
   Kubik (x hoch 3) *)
function Kub (x : double) : double;
begin
  Kub:=sqr(x)*x;
  end;

(* ----------FUNCTION--TAN---------------------------IEAP-KIEL----------
  Tangens
  *)
function Tan (x : double) : double;
var
  c : double;
begin
  c:=cos(x);
  if c<>0.0 then Tan:=sin(x)/c
  else Tan:=Maxdouble;
  end;

(* ----------FUNCTION--ArcTan1-----------------------IEAP-KIEL----------
   ArcTan mit Quadrantenbestimmung, Winkel von 0 .. 360 Grad
   linksweisend gegen Ost (math. pos. Sinn) *)
function ArcTan1 (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=90.0-ArcTan (x/y)/Pi180;
    if y<0.0 then z:=z+180.0;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=ArcTan (y/x)/Pi180;
    if x<0.0 then z:=z+180.0;
    end;
  if z<0.0 then z:=z+360.0;
  ArcTan1:=z;
  end;

(* ----------FUNCTION--ArcTan2-----------------------IEAP-KIEL----------
   ArcTan mit Quadrantenbestimmung, Winkel von 0 .. 360 Grad
   rechtsweisend gegen Nord
*)
function ArcTan2 (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=ArcTan (x/y)/Pi180;
    if y<0.0 then z:=z+180.0
    else if x<0.0 then z:=z+360.0;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=90.0-ArcTan (y/x)/Pi180;
    if x<0.0 then z:=z+180.0;
    end;
  ArcTan2:=z;
  end;

(* ----------FUNCTION--ArcTan-----------------------IEAP-KIEL----------
   ArcTan mit Quadrantenbestimmung, Winkel in Bogenma
   +/- Pi im math. pos. Sinn *)
function ArcTanB (x,y : double) : double;
var
  z : double;
begin
  if abs(x)<abs(y) then begin
    z:=0.5*Pi-ArcTan (x/y);
    if y<0.0 then z:=z+Pi
    else if x<0.0 then z:=z+2*Pi;
    end
  else begin
    if x=0.0 then z:=0.0
    else z:=ArcTan (y/x);
    if x<0.0 then z:=z+Pi;
    end;
  if z<=-Pi then z:=z+2*Pi;
  if z>Pi then z:=z-2*Pi;
  ArcTanB:=z;
  end;

(* ----------FUNCTION--RMIN--------------------------IEAP-KIEL----------
   Mininum zweier Realzahlen
*)
function rmin (x,y : double) : double;
begin
  if x<y then rmin:=x else rmin:=y;
  end;

(* ----------FUNCTION--RMAX--------------------------IEAP-KIEL----------
   Maximum zweier Realzahlen
*)
function rmax (x,y : double) : double;
begin
  if x<y then rmax:=y else rmax:=x;
  end;

// Anzahl der Stellen einer Integerzahl
function NumberOfDigits (n : integer) : integer;
begin
  Result:=length(Trim(IntToStr(n)));
  end;

(* ----------PROCEDURE-CheckValue---------------------IEAP-KIEL---------- *)
function CheckValue (s     : string) : boolean;
begin
  Result:=true;
  if length(s)>0 then begin
    try
      StrToFloat(s);
    except
      on EConvertError do Result:=false;
      end;
    end;
  end;

(* ----------FUNCTION DECIMALROUND------------------------IEAP-KIEL---------- *)
(* an Dezimalstelle runden *)
function DecimalRound (Value : double; Decimal : integer) : double;
var
  v : integer;
begin
  if Decimal<-2 then Decimal:=-2;
  if Decimal>8 then Decimal:=8;
  v:=round(abs(Value)*FFak[Decimal]);
  Result:=v/FFak[Decimal];
  if Value<0 then Result:=-Result;
  end;

(* ----------PROCEDURE-NORMAL------------------------IEAP-KIEL---------- *)
(* Normalisieren eines Wertes in den Bereich 0.1 .. 999.9 *)
procedure Normal (var Value,Factor : double;
                  var PrefixIndex  : integer);
var
  AVal : double;
  k    : integer;
begin
  AVal:=abs(Value);
  if (AVal<1E-15) or (AVal>1E15) then k:=6
  else begin
    k:=1;
    if (AVal>=0.999E-13) then inc(k);
    if (AVal>=0.999E-10) then inc(k);
    if (AVal>=0.999E-7) then inc(k);
    if (AVal>=0.999E-4) then inc(k);
    if (AVal>=0.999E-1) then inc(k);
    if (AVal>=0.999E3) then inc(k);
    if (AVal>=0.999E6) then inc(k);
    if (AVal>=0.999E9) then inc(k);
    if (AVal>=0.999E12) then inc(k);
    end;
  Factor:=PfFak[k];
  Value:=Value*Factor;
  PrefixIndex:=k;
  end;

(* ----------FUNCTION--PREFIXVAL---------------------IEAP-KIEL---------- *)
(* Zahlenwert mit nachfolgendem Einheitenvorsatz erzeugen *)
function PrefixVal (Value   : double;
                    n,nd    : integer;
                    Einh    : EinhStr) : string;
(* n    = Anzahl der Stellen
        = 0: automatische Stellenanpassung
   nd   = Anzahl der Stellen hinter dem Komma
          <0 : automatisch max. -(nd+1) Stellen hinter dem Komma
   Einh = nachfolgendes Einheitenzeichen *)
var
  s    : string;
  pf   : char;
  f    : double;
  i    : integer;
begin
  Normal (Value,f,i); pf:=Pfx[i];
  if nd<0 then begin
    nd:=abs(nd)-2;
    if abs(Value)<100.0 then inc(nd);
    if abs(Value)<10.0 then inc(nd);
    if abs(Value)<1.0 then inc(nd);
    if abs(Value)<0.0999 then nd:=0;
    if nd<0 then nd:=0;
    end;
  if n=0 then begin
    n:=3+nd; if nd>0 then inc(n);
    if abs(Value)<100.0 then dec(n);
    if abs(Value)<10.0 then dec(n);
    end;
  if Value<0.0 then inc(n);
  s:=FloatToStrF(Value,ffFixed,n,nd);
  if Einh='' then begin
    if pf<>' ' then PrefixVal:=s+pf
    else PrefixVal:=s;
    end
  else if pf<>' ' then PrefixVal:=s+' '+pf+Einh
    else PrefixVal:=s+' '+Einh+' ';
  end;

(* ----------FUNCTION--VALPREFIX---------------------IEAP-KIEL---------- *)
(* Zahlenwert aus String mit nachfolgendem Einheitenvorsatz erzeugen *)
function ValPrefix (ValStr  : string;
                    var Value : double) : boolean;
(* PrefixVal = true : ok
             = false : Fehler bei der Konvertierung *)
var
  k,n  : integer;
  c    : char;
begin
  ValStr:=Trim(ValStr);
  (* Prfe letztes Zeichen *)
  n:=length(ValStr);
  c:=ValStr[n]; Value:=0.0; k:=0;
  if (c<'0') or (c>'9') then begin
    k:=1;
    while (k<=PfAnz) and (Pfx[k]<>c) do inc(k);
    end;
  if k<=PfAnz then begin
    if k>0 then begin
      dec(n);
      while (n>0) and (ValStr[n]=' ') do dec(n);
      end;
    try
      Value:=StrToFloat(copy(ValStr,1,n));
      Value:=Value/PfFak[k];
      ValPrefix:=true;
    except
      ValPrefix:=false;
      end;
    end
  else ValPrefix:=false;
  end;

(* ----------FUNCTION--FloatToStringP---------------------IEAP-KIEL---------- *)
procedure FloatToEng (Value   : double; DecSep  : char; n : integer;
                      var Mant : string; var Exp : integer);
// Value   = Wert
// n       = Anzahl der signifikanten Stellen
// DecSep  = Dez.trenner
// Mant    = Mantisse mit n signif. Stellen
// Exp     = Exponent in 3er-Stufen
var
  f    : double;
  nd,v,w : integer;
begin
  Normal (Value,f,Exp);
  if n<=1 then n:=2;
  if n>8 then n:=8;
  nd:=0;
  if abs(Value)>=0.99999 then inc(nd);
  if abs(Value)>=9.9999 then inc(nd);
  if abs(Value)>=99.999 then inc(nd);
  f:=abs(Value)*FFak[n-nd]+1E-6;
  v:=round(f); w:=trunc(f);
  if NumberOfDigits(v)>NumberOfDigits(w) then begin
    v:=v div 10; inc(nd);
    end;
  if n<nd then v:=v*NFak[nd-n];
  Mant:=IntToStr(v);
  if nd>0 then begin
    if n>nd then insert(DecSep,Mant,nd+1);
    end
  else Insert('0'+DecSep,Mant,1);
  if Value<0 then Insert('-',Mant,1);
  end;

(* Zahlen auf "Eng-Format" umsetzen (3er Potenzen) *)
function FloatToStrP (Value   : double;
                      DecSep  : char;
                      n       : integer) : string;
(* n    = Anzahl der signifikanten Stellen *)
var
  s : string;
  k : integer;
begin
  if n=0 then Result:=IntToStr(round(Value))
  else begin
    FloatToEng (Value,DecSep,n,s,k);
    if k<>6 then Result:=s+'E'+IntToStr((k-6)*3)
    else Result:=s;
    end;
  end;

(* ----------FUNCTION--FloatToPlotString---------------------IEAP-KIEL---------- *)
(* Zahlen auf "Eng-Format" mit 10^..  umsetzen *)
function FloatToPlotStr (Value   : double;
                         n       : integer) : string;
(* n    = Anzahl der signifikanten Stellen *)
var
  s : string;
  i : integer;
begin
  s:=FloatToStrP(Value,DecimalSeparator,n);
  i:=Pos('E',s);
  if i>0 then begin
    delete(s,i,1); insert('10^',s,i); s:=s+'_';
    end;
  Result:=s;
  end;

(* ----------FUNCTION--FloatToPlotString---------------------IEAP-KIEL---------- *)
(* Zahlen auf "Eng-Format" mit Vorstzen (m,k,..)  umsetzen *)
function FloatToPrefixStr (Value     : double;
                           n         : integer;
                           Separator : string) : string;
(* n    = Anzahl der signifikanten Stellen *)
var
  s   : string;
  k   : integer;
begin
  FloatToEng (Value,DecimalSeparator,n,s,k);
  if (k<1) or (k>PfAnz) then Result:=s+'E'+IntToStr((k-6)*3)
  else if k<>6 then Result:=s+Separator+pfx[k] else Result:=s;
  end;

(* ----------FUNCTION--FloatToFixString---------------------IEAP-KIEL---------- *)
(* Zahlen auf "ffXixed-Format" umsetzen *)
function FloatToFixStr (Value     : double;
                        n         : integer) : string;
(* n    = Anzahl der signifikanten Stellen + 100 * Stellen hinter dem Komma *)
var
  k   : integer;
begin
  k:=n div 100;
  n:=n mod 100;
  if n<k+2 then n:=k+2;
  Result:=FloatToStrF (Value,ffFixed,n,k);
  end;

(* ----------Initialisierung-------------------------IEAP-KIEL---------- *)
begin
  (* Initialisierung der Konstanten *)
  ln10:=ln(10); ln2:=ln(2); sqrt2:=sqrt(2);
  end.
