(* Delphi Unit
   verschiedene Unterroutinen fr Windows Netzwerkverbindungen
   ===========================================================
   
    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Sep. 2002
     letzte nderung: Feb. 2008
   *)

unit WinNet;

interface

uses
  Windows, Classes, SysUtils, Controls;

type
  TUserAccount = record
    Username,Password : string;
    end;

  TDriveType = (dtUnknown,dtNoRoot,dtRemovable,dtFixed,dtRemote,dtCdRom,dtRamDisk);
  TDriveTypes = set of TDriveType;
  TPathType =  (ptNotAvailable,ptAvailable,ptRelative,ptRemovable,ptRemote);

  TDriveProperties = class(TObject)
    Number    : integer;
    DriveType : TDriveType;
    DriveName,
    VolName   : string
    end;

const
  DriveTypeList : array[TDriveType] of string =
                ('unknown','no root','removable','fixed','remote','CD-ROM','Ram-Disk');

// Typ eines Laufwerkes ermitteln
function DriveType (const Path : string) : TDriveType;

// Typ eines Pfades ermitteln (siehe TPathtype)
function CheckPath (const Path : string) : TPathType;

// Prfe Laufwerk auf Verfgbarkeit
function CheckForDriveAvailable (const Path : string; var VolumeID : string) : boolean;

// Liste aller Laufwerke der Typen "UseTypes" aufbauen
procedure BuildDriveList(DriveList : TStrings; UseTypes : TDriveTypes);

// Zu einem Volume gehrendes Laufwerk ermitteln
function GetDriveForVolume (const Vol : string; FirstDrive : integer) : string;

// Name des angemeldeten Benutzers ermitteln
function GetUserName : string;

// Prfen, ob ein in einem Pfad enthaltenes Laufwerk (lokal oder Netz) verfgbar ist.
// StdUser: Name des angemeldeten Benutzers oder leer
function NetPathAvailable (const Path,StdUser : string; Prompt : boolean) : boolean;

function ReconnectPath (const Path : string) : boolean;

// Account eines alternativen Benutzers abfragen
function ReadAltUserAccount (NetResource : string) : boolean;

// Alternativen Benutzernamen setzen
procedure SetAltUserAccount (User,Pwd : string);

// Alternativen Benutzernamen lschen
procedure ResetAltUserAccount;

function GetVolumeNameForVolumeMountPoint(lpszVolumeMountPoint: PChar;
  lpszVolumeName : PChar; var cchBufferLength : DWord): BOOL; stdcall;
{$EXTERNALSYM GetVolumeNameForVolumeMountPoint}

implementation

uses LogonDlg, JrUtils, GnuGetText;

function GetVolumeNameForVolumeMountPoint; external 'kernel32.dll' name 'GetVolumeNameForVolumeMountPointA';

// Typ eines Laufwerkes ermitteln
function DriveType (const Path : string) : TDriveType;
begin
  case GetDriveType(pchar(ExtractFileDrive(Path))) of
  DRIVE_NO_ROOT_DIR : Result:=dtNoRoot;
  DRIVE_REMOVABLE   : Result:=dtRemovable;
  DRIVE_FIXED       : Result:=dtFixed;
  DRIVE_REMOTE      : Result:=dtRemote;
  DRIVE_CDROM       : Result:=dtCdRom;
  DRIVE_RAMDISK     : Result:=dtRamDisk;
  else Result:=dtUnknown;
    end;
  end;

// Typ eines Pfades ermitteln (siehe TPathtype)
function CheckPath (const Path : string) : TPathType;
var
  dr : string;
  dt : TDriveType;
begin
  dr:=ExtractFileDrive(IncludeTrailingPathDelimiter(Path));
  if length(dr)>0 then begin
    if (copy(dr,1,2)='\\') then Result:=ptRemote    // Netzwerkumgebung
    else begin                                      // Pfad mit Laufwerksangabe
      dt:=DriveType(dr);
      case dt of
      dtRemote : Result:=ptRemote;          // Netzlaufwerk
      dtNoRoot : Result:=ptNotAvailable;    // nicht verfgbar
      dtCdRom,
      dtRemovable : Result:=ptRemovable;    // Laufwerk mit Wechselmedium
      else Result:=ptAvailable;
        end;
      end;
    end
  else Result:=ptRelative;
  end;

// Prfe Laufwerk auf Verfgbarkeit
function CheckForDriveAvailable (const Path : string; var VolumeID : string) : boolean;
var
  v : pchar;
  d : string;
  n,cl,sf : dword;
begin
  d:=IncludeTrailingPathDelimiter(ExtractFileDrive(Path));
  n:=50; v:=StrAlloc(n);
  Result:=GetVolumeInformation(pchar(d),v,n,nil,cl,sf,nil,0);
  if Result then VolumeID:=v else VolumeID:=dgettext('units','not available');
  end;

// Liste aller Laufwerke der Typen "UseTypes" aufbauen
procedure BuildDriveList (DriveList : TStrings; UseTypes : TDriveTypes);
var
  i         : integer;
  DriveBits : set of 0..25;
  dp        : TDriveProperties;
begin
  DriveList.Clear;
  Integer(DriveBits):=GetLogicalDrives;
  for i:=0 to 25 do begin
    if not (i in DriveBits) then Continue;
    dp:=TDriveProperties.Create;
    with dp do begin
      Number:=i;
      DriveName:=Char(i+Ord('A'))+':\';
      DriveType:=TDriveType(GetDriveType(PChar(DriveName)));
      CheckForDriveAvailable(DriveName,VolName);
      end;
    if dp.DriveType in UseTypes then DriveList.AddObject(dp.VolName,dp)
    else dp.Free;;
    end;
  end;

// Zu einem Volume gehrendes Laufwerk ermitteln
function GetDriveForVolume (const Vol : string; FirstDrive : integer) : string;
var
  i         : integer;
  DriveBits : set of 0..25;
  sd,sv     : string;
begin
  Result:='';
  Integer(DriveBits):=GetLogicalDrives;
  for i:=FirstDrive to 25 do begin
    if not (i in DriveBits) then Continue;
    sd:=Char(i+Ord('A'))+':\';
    if CheckForDriveAvailable(sd,sv) and AnsiSameText(sv,Vol) then begin
      Result:=sd; Exit;
      end;
    end;
  end;

const
  NwTestName = 'test.tmp';

var
  AltUserAccount : TUserAccount;

// Name des angemeldeten Benutzers ermitteln
function GetUserName : string;
var
  ul    : dword;
  un    : pchar;
begin
  ul:=1024; GetMem (un,ul);
  if WNetGetUser(nil,un,ul)=NO_ERROR then Result:=un else Result:='';
  FreeMem (un);
  end;

// Prfen, ob ein in einem Pfad enthaltenes Laufwerk (lokal oder Netz) verfgbar ist.
// Bei Netzlaufwerken wird geprft, ob die Verbindung zum Schreiben verfgbar ,
// wenn nicht, wird versucht, sie herzustellen. Dazu wird zuerst versucht,
// dies mit dem angemeldeten Benutzer zu machen, dann mit einem eingetragenen
// alternativen Benutzer ("AltUserAccount")
// Prompt = true:  Wenn das nicht geht, wird ein Benutzername und ein Passwort abgefragt.
function NetPathAvailable (const Path,StdUser : string; Prompt : boolean) : boolean;
var
  User,Pwd : string;
  ok,ende  : boolean;
  ec,nl    : dword;
  NetRes   : TNetResource;
  nn       : pchar;

  function CheckPathWritable (const Path : string) : boolean;
  var
    fsT      : TextFile;
    s        : string;
  begin
  // versuche, in Nw-Pfad zu schreiben
    if not DirectoryExists(Path) then Result:=ForceDirectories(Path)
    else Result:=true;
    if Result then begin
      s:=IncludeTrailingPathDelimiter(Path)+NwTestName;
      AssignFile (fsT,s);
      {$I-} Rewrite(fsT);  {$I+}
      Result:=IoResult=0;
      if Result Then begin
        CloseFile(fsT);
        DeleteFile(s);
        end;
      end;
    end;

begin
  if not CheckPathWritable (Path) then begin
    nl:=1024; nn:=StrAlloc(nl);
    with NetRes do begin
      dwScope:=0; dwDisplayType:=0;
      lpProvider:=''; lpLocalName:='';
      dwUsage:=RESOURCEUSAGE_CONNECTABLE;
      dwType:=RESOURCETYPE_DISK;
      if (copy(Path,1,2)='\\') then begin  // Netzwerkumgebung
        lpRemoteName:=pchar(ExcludeTrailingPathDelimiter(Path));
        end
      else begin
        lpLocalName:=pchar(ExtractFileDrive(Path));
        WNetGetConnection(lpLocalName,nn,nl);   // Netzwerkname des Pfads
        lpRemoteName:=nn;
        end;
      end;
    // als angemeldeter Benutzer versuchen
    if length(StdUser)>0 then ec:=WNetAddConnection2(NetRes,nil,pchar(StdUser),0)
    else ec:=ERROR_LOGON_FAILURE;
    if (ec=NO_ERROR) and not CheckPathWritable (NetRes.lpRemoteName) then ec:=ERROR_SESSION_CREDENTIAL_CONFLICT;
    // Benutzer und Passwort wie im Dialog zuvor angegeben
    with AltUserAccount do if (ec<>NO_ERROR) and (length(Username)>0) then begin
      ec:=WNetCancelConnection2(NetRes.lpRemoteName,0,false);
      if (ec=NO_ERROR) or (ec=ERROR_NOT_CONNECTED) then begin
            ec:=WNetAddConnection2(NetRes,pchar(Password),pchar(Username),0);
        end;
      end;
    if Prompt and ((ec=ERROR_SESSION_CREDENTIAL_CONFLICT) or (ec=ERROR_LOGON_FAILURE)) then begin
      Ende:=false;
      repeat
        if ec=ERROR_SESSION_CREDENTIAL_CONFLICT then
          ec:=WNetCancelConnection2(NetRes.lpRemoteName,0,false)
        else ec:=NO_ERROR;
        if (ec=NO_ERROR) or (ec=ERROR_NOT_CONNECTED) then begin
          Pwd:='';
          ec:=WNetAddConnection2(NetRes,pchar(Pwd),pchar(User),CONNECT_INTERACTIVE or CONNECT_PROMPT);
          Ende:=ec=NO_ERROR;
          end
        else ec:=ERROR_CANCELLED;
        until Ende or (ec=ERROR_CANCELLED);
      ok:=Ende;
      end
    else ok:=ec=NO_ERROR;
    StrDispose(nn);
    Result:=ok;
    end
  else Result:=true;
  end;

function ReconnectPath (const Path : string) : boolean;
var
  nl       : dword;
  NetRes   : TNetResource;
  nn       : pchar;
begin
  nl:=1024; nn:=StrAlloc(nl);
  with NetRes do begin
    dwScope:=0; dwDisplayType:=0;
    lpProvider:=''; lpLocalName:='';
    dwUsage:=RESOURCEUSAGE_CONNECTABLE;
    dwType:=RESOURCETYPE_DISK;
    if (copy(Path,1,2)='\\') then begin  // Netzwerkumgebung
      lpRemoteName:=pchar(ExcludeTrailingPathDelimiter(Path));
      end
    else begin
      lpLocalName:=pchar(ExtractFileDrive(Path));
      WNetGetConnection(lpLocalName,nn,nl);   // Netzwerkname des Pfads
      lpRemoteName:=nn;
      end;
    end;
  // als angemeldeter Benutzer versuchen
  Result:=WNetAddConnection2(NetRes,nil,nil,0)=NO_ERROR;
  end;

// Account eines alternativen Benutzers abfragen
function ReadAltUserAccount (NetResource : string) : boolean;
var
  User,Pwd : string;
begin
  User:=''; Pwd:='';
  if InputUserAccount(dgettext('units','Connect to ..'),NetResource,false,User,Pwd)=mrOK then
       with AltUserAccount do begin
    Username:=User; Password:=Pwd;
    Result:=true;
    end
  else Result:=false;
  end;

// Alternativen Benutzernamen setzen
procedure SetAltUserAccount (User,Pwd : string);
begin
  with AltUserAccount do begin
    Username:=User; Password:=Pwd;
    end
  end;

// Alternativen Benutzernamen lschen
procedure ResetAltUserAccount;
begin
  with AltUserAccount do begin
    Username:=''; Password:='';
    end;
  end;

begin
  ResetAltUserAccount;
  end.
